from enum import Enum

from pydantic import BaseModel, ConfigDict


class CatPersonality(Enum):
    CUTE = "CUTE"
    EVIL = "EVIL"


class Cat(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    id: int
    name: str
    breed: str
    age: float
    personality: CatPersonality



from abc import abstractmethod

from examples.fastapi.routes.cats.domain import Cat


class CatRepository:
    @abstractmethod
    def get(self, cat_id: int) -> Cat: ...

    @abstractmethod
    def set(self, cat: Cat) -> None: ...

    @abstractmethod
    def next_id(self) -> int: ...



import asyncio
from contextlib import asynccontextmanager

from starlette.applications import Starlette

from engin import Engin, Invoke, Lifecycle, Provide
from engin.extensions.asgi import ASGIEngin, ASGIType


def a(lifecycle: Lifecycle) -> None:
    @asynccontextmanager
    async def _raise_err() -> None:
        yield
        raise RuntimeError("Error in Shutdown!")

    lifecycle.append(_raise_err())


B_LIFECYCLE_STATE = 0


def b(lifecycle: Lifecycle) -> None:
    @asynccontextmanager
    async def _b_startup() -> None:
        global B_LIFECYCLE_STATE
        B_LIFECYCLE_STATE = 1
        yield
        B_LIFECYCLE_STATE = 2

    lifecycle.append(_b_startup())


async def test_error_in_shutdown_when_start():
    engin = Engin(Invoke(a), Invoke(b))

    await asyncio.wait_for(engin.start(), timeout=0.5)
    assert B_LIFECYCLE_STATE == 1

    await asyncio.wait_for(engin.stop(), timeout=0.5)
    assert B_LIFECYCLE_STATE == 2


async def test_error_in_shutdown_when_run():
    engin = Engin(Invoke(a), Invoke(b))

    await asyncio.wait_for(engin.run(), timeout=0.5)
    assert B_LIFECYCLE_STATE == 2


async def test_error_in_shutdown_asgi():
    def asgi_type() -> ASGIType:
        return Starlette()

    engin = ASGIEngin(Invoke(a), Invoke(b), Provide(asgi_type))

    await engin.start()
    assert B_LIFECYCLE_STATE == 1

    await engin.stop()
    assert B_LIFECYCLE_STATE == 2



import asyncio

import pytest


@pytest.fixture
async def aio_benchmark(benchmark):
    async def run_async_coroutine(func, *args, **kwargs):
        return await func(*args, **kwargs)

    def _wrapper(func, *args, **kwargs):
        if asyncio.iscoroutinefunction(func):

            @benchmark
            def _():
                future = asyncio.ensure_future(run_async_coroutine(func, *args, **kwargs))
                return asyncio.get_event_loop().run_until_complete(future)
        else:
            benchmark(func, *args, **kwargs)

    return _wrapper



import asyncio
import logging

from engin import Engin, Entrypoint, Provide
from examples.tutorial.consumer import Consumer, consumer_factory
from examples.tutorial.publisher import Publisher, publisher_factory
from examples.tutorial.valkey_client import ValkeyBlock

logging.basicConfig(level=logging.INFO)

engin = Engin(
    ValkeyBlock(),
    Provide(publisher_factory),
    Provide(consumer_factory),
    Entrypoint(Publisher),
    Entrypoint(Consumer),
)


if __name__ == "__main__":
    asyncio.run(engin.run())



from typing import Annotated
from unittest.mock import Mock

import pytest

from engin import Entrypoint, Invoke, Provide, Supply
from tests.deps import int_provider, make_aliased_int, make_int


def test_provide_discriminates_singular():
    provider = int_provider()
    assert not provider.is_multiprovider
    assert provider.return_type_id.type is int
    assert not provider.return_type_id.multi


def test_provide_discriminates_multi():
    def i_provide_multi() -> list[int]:
        return [3]

    provider = Provide(i_provide_multi)
    assert provider.is_multiprovider
    assert provider.return_type_id.type is int
    assert provider.return_type_id.multi


def test_supply_singular():
    class Foo: ...

    supply = Supply(Foo())
    assert not supply.is_multiprovider
    assert supply.return_type_id.type is Foo
    assert not supply.return_type_id.multi


def test_supply_multi():
    class Foo: ...

    supply = Supply([Foo()])
    assert supply.return_type_id.type is Foo
    assert supply.return_type_id.multi
    assert supply.is_multiprovider


def test_provide_with_alias():
    provider = Provide(make_aliased_int)

    assert provider.return_type_id.type


def test_provide_with_annotation():
    def make_str_1() -> Annotated[str, "1"]:
        return "bar"

    provider = Provide(make_str_1)

    assert provider.return_type_id.type
    assert str(provider.return_type_id) == "Annotated[str, 1]"


def test_dependency_sources():
    provide = Provide(make_int)
    assert provide.source_module == "tests.test_dependencies"
    assert provide.source_package == "tests"

    supply = Supply(3)
    assert supply.source_module == "tests.test_dependencies"
    assert supply.source_package == "tests"

    invoke = Invoke(make_int)
    assert invoke.source_module == "tests.test_dependencies"
    assert invoke.source_package == "tests"

    entrypoint = Entrypoint(3)
    assert entrypoint.source_module == "tests.test_dependencies"
    assert entrypoint.source_package == "tests"


def test_provider_cannot_depend_on_self():
    def invalid_provider_1(a: int) -> int:
        return 1

    def invalid_provider_2(a: list[int]) -> list[int]:
        return [1]

    with pytest.raises(ValueError, match="return type"):
        Provide(invalid_provider_1)

    with pytest.raises(ValueError, match="return type"):
        Provide(invalid_provider_2)


def test_provides_implicit_overrides_providers():
    provide_a = int_provider()
    provide_b = int_provider()

    engin = Mock()
    engin._providers = {}

    provide_a.apply(engin)

    with pytest.raises(RuntimeError, match="implicit"):
        provide_b.apply(engin)


def test_provides_implicit_overrides_supply():
    provide_a = Supply(3)
    provide_b = Supply(4)

    engin = Mock()
    engin._providers = {}

    provide_a.apply(engin)

    with pytest.raises(RuntimeError, match="implicit"):
        provide_b.apply(engin)


def test_provides_explicit_overrides_allowed():
    provide_a = int_provider()
    provide_b = int_provider(override=True)

    engin = Mock()
    engin._providers = {}

    provide_a.apply(engin)
    provide_b.apply(engin)


def test_provides_implicit_overrides_allowed_when_3rd_party():
    provide_a = int_provider()
    provide_b = int_provider()

    provide_a._source_package = "foo"

    engin = Mock()
    engin._providers = {}

    provide_a.apply(engin)
    provide_b.apply(engin)


def test_provide_as_type():
    provide = int_provider(as_type=float)
    assert provide.return_type is float
    assert provide.signature.return_annotation is float


def test_supply_as_type():
    supply = Supply(3, as_type=float)
    assert supply.return_type is float
    assert supply.signature.return_annotation is float



import logging

from valkey import ResponseError
from valkey.asyncio import Valkey

from engin import Supervisor


class Consumer:
    def __init__(self, valkey: Valkey) -> None:
        self._valkey = valkey
        self._total = 0

    async def run(self) -> None:
        logging.info("Consumer starting")
        try:
            await self._valkey.xgroup_create("numbers", "total", mkstream=True)
        except ResponseError:
            pass  # group already exists

        while True:
            messages = await self._valkey.xreadgroup("total", "consumer", {"numbers": ">"})
            for _, stream_messages in messages:
                for _, message in stream_messages:
                    number = int(message[b"number"])
                    self._total += number
                    logging.info(f"Consumed: {number}, Total: {self._total}")


def consumer_factory(valkey: Valkey, supervisor: Supervisor) -> Consumer:
    consumer = Consumer(valkey=valkey)

    # run the consumer as a supervised application task
    supervisor.supervise(consumer.run)

    return consumer



from fastapi import APIRouter
from pytest_mock import MockerFixture
from typer.testing import CliRunner

from engin import Engin, Entrypoint, Invoke, Supply
from engin._cli._graph import cli
from tests.deps import ABlock


def invoke_something(route: APIRouter) -> None: ...


api_router = APIRouter()


@api_router.get("/hello")
def get_hello() -> str:
    return "hello"


engin = Engin(
    ABlock,
    Supply(["3"]),
    Supply(api_router),
    Invoke(invoke_something),
    Entrypoint(list[str]),
)
runner = CliRunner()


def test_cli_graph(mocker: MockerFixture) -> None:
    mocker.patch("engin._cli._graph.wait_for_interrupt", side_effect=KeyboardInterrupt)
    result = runner.invoke(app=cli, args=["tests.cli.test_graph:engin"])
    assert result.exit_code == 0


def test_cli_invalid_app_path() -> None:
    result = runner.invoke(app=cli, args=["tests.cli.foo"])
    assert result.exit_code == 1
    assert "module" in result.output


def test_cli_invalid_app_path_2() -> None:
    result = runner.invoke(app=cli, args=["tests.cli.foo:engin"])
    assert result.exit_code == 1
    assert "module" in result.output


def test_cli_invalid_app_attribute() -> None:
    result = runner.invoke(app=cli, args=["tests.cli.test_graph:foo"])
    assert result.exit_code == 1
    assert "no attribute" in result.output


def test_cli_invalid_app_instance() -> None:
    result = runner.invoke(app=cli, args=["tests.cli.test_graph:runner"])
    assert result.exit_code == 1
    assert "Engin" in result.output



from typer.testing import CliRunner

from engin import Engin
from engin._cli._inspect import cli
from tests.deps import ABlock

engin = Engin(ABlock)
runner = CliRunner()


def test_cli_inspect() -> None:
    result = runner.invoke(
        app=cli,
        args=["tests.cli.test_inspect:engin", "--type", "float[]", "--verbose"],
    )
    assert result.exit_code == 0



import inspect
import logging
import typing
from collections.abc import Awaitable, Callable
from dataclasses import dataclass
from enum import Enum
from types import TracebackType
from typing import TypeAlias, assert_never

import anyio
from anyio import get_cancelled_exc_class

if typing.TYPE_CHECKING:
    from anyio.abc import TaskGroup

LOG = logging.getLogger("engin")

TaskFactory: TypeAlias = Callable[[], Awaitable[None]]


class OnException(Enum):
    SHUTDOWN = 0
    """
    Cancel all other supervised tasks and shutdown the Engin.
    """

    RETRY = 1
    """
    Retry the task.
    """

    IGNORE = 2
    """
    The task will be not be retried and the engin will not be stopped, other tasks will
    continue to run.
    """


@dataclass(kw_only=True, slots=True, eq=False)
class _SupervisorTask:
    """
    Attributes:
        - factory: a coroutine function that can create the task.
        - on_exception: determines the behaviour when task raises an exception.
        - complete: will be set to true if task stops for any reason except cancellation.
        - last_exception: the last exception raised by the task.
    """

    factory: TaskFactory
    on_exception: OnException
    complete: bool = False
    last_exception: Exception | None = None

    async def __call__(self) -> None:
        # loop to allow for restarting erroring tasks
        while True:
            try:
                await self.factory()
                self.complete = True
                return
            except get_cancelled_exc_class() as err:
                LOG.debug(f"supervised task '{self.name}' was cancelled", exc_info=err)
                raise
            except Exception as err:
                self.last_exception = err
                if self.on_exception == OnException.IGNORE:
                    LOG.warning(
                        f"supervisor task '{self.name}' raised {type(err).__name__} "
                        "which will be ignored",
                        exc_info=err,
                    )
                    self.complete = True
                    return
                if self.on_exception == OnException.RETRY:
                    LOG.warning(
                        f"supervisor task '{self.name}' raised {type(err).__name__} "
                        "which will be retried",
                        exc_info=err,
                    )
                    continue
                if self.on_exception == OnException.SHUTDOWN:
                    LOG.error(
                        f"supervisor task '{self.name}' raised {type(err).__name__}, "
                        "starting shutdown",
                        exc_info=err,
                    )
                    self.complete = True
                    raise get_cancelled_exc_class() from None
                assert_never(self.on_exception)

    @property
    def name(self) -> str:
        factory = self.factory
        if inspect.ismethod(factory):
            return f"{factory.__self__.__class__.__name__}.{factory.__func__.__name__}"
        if inspect.isclass(factory):
            return type(factory).__name__
        if inspect.isfunction(factory):
            return factory.__name__
        return str(factory)


class Supervisor:
    def __init__(self) -> None:
        self._tasks: list[_SupervisorTask] = []
        self._task_group: TaskGroup | None = None

    def supervise(
        self, func: TaskFactory, *, on_exception: OnException = OnException.SHUTDOWN
    ) -> None:
        self._tasks.append(_SupervisorTask(factory=func, on_exception=on_exception))

    @property
    def empty(self) -> bool:
        return not self._tasks

    async def __aenter__(self) -> None:
        if not self._tasks:
            return

        self._task_group = await anyio.create_task_group().__aenter__()

        for task in self._tasks:
            LOG.info(f"supervising task: {task.name}")
            self._task_group.start_soon(task, name=task.name)

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
        /,
    ) -> None:
        if self._task_group:
            if not self._task_group.cancel_scope.cancel_called:
                self._task_group.cancel_scope.cancel()
            await self._task_group.__aexit__(exc_type, exc_value, traceback)



import asyncio
import contextlib
from asyncio import CancelledError
from dataclasses import dataclass

from engin import OnException, Supervisor


async def delayed_error_task():
    await asyncio.sleep(0.05)
    raise RuntimeError("Process errored")


async def happy_task():
    await asyncio.sleep(1)


async def endless_task():
    await asyncio.sleep(999)


async def block_task():
    await asyncio.sleep(999)


@dataclass
class ClassHappyTask:
    async def run(self) -> None:
        await asyncio.sleep(1)


async def test_empty_supervisor():
    supervisor = Supervisor()

    with contextlib.suppress(CancelledError):
        async with supervisor:
            await asyncio.sleep(0.1)


async def test_supervisor_on_exception_retry():
    supervisor = Supervisor()
    attempt = 0

    async def retry_task():
        nonlocal attempt
        if attempt == 0:
            attempt += 1
            raise RuntimeError("Process errored")

    supervisor.supervise(retry_task, on_exception=OnException.RETRY)

    with contextlib.suppress(CancelledError):
        async with supervisor:
            await asyncio.sleep(0.1)

    assert supervisor._tasks[0].complete
    assert isinstance(supervisor._tasks[0].last_exception, RuntimeError)
    assert attempt == 1


async def test_supervisor_on_exception_ignore():
    supervisor = Supervisor()

    async def error_task():
        raise RuntimeError("Process errored")

    async def complete_task():
        await asyncio.sleep(0.09)

    supervisor.supervise(error_task, on_exception=OnException.IGNORE)
    supervisor.supervise(complete_task, on_exception=OnException.SHUTDOWN)

    with contextlib.suppress(CancelledError):
        async with supervisor:
            await asyncio.sleep(0.1)

    assert supervisor._tasks[0].complete
    assert isinstance(supervisor._tasks[0].last_exception, RuntimeError)
    assert supervisor._tasks[1].complete
    assert supervisor._tasks[1].last_exception is None


async def test_supervisor():
    supervisor = Supervisor()

    supervisor.supervise(delayed_error_task)
    supervisor.supervise(happy_task)
    supervisor.supervise(ClassHappyTask().run)

    with contextlib.suppress(CancelledError):
        async with supervisor:
            await asyncio.sleep(1)

    # task one completed and has error (as raised exception)
    assert supervisor._tasks[0].complete
    assert supervisor._tasks[0].last_exception is not None

    # task two did not complete and has no error (as was cancelled)
    assert not supervisor._tasks[1].complete
    assert supervisor._tasks[1].last_exception is None

    # task three did not complete and has no error (as was cancelled)
    assert not supervisor._tasks[2].complete
    assert supervisor._tasks[2].last_exception is None


async def test_supervisor_is_cancellable():
    supervisor = Supervisor()

    supervisor.supervise(endless_task)

    await supervisor.__aenter__()
    await asyncio.wait_for(supervisor.__aexit__(None, None, None), 0.1)



from collections.abc import Callable
from typing import Annotated, TypeAlias

from engin import TypeId


def test_type_id_of_int():
    type_id = TypeId.from_type(int)
    assert type_id == TypeId(type=int, multi=False)
    assert str(type_id) == "int"


def test_type_id_of_list_of_int():
    type_id = TypeId.from_type(list[int])
    assert type_id == TypeId(type=int, multi=True)
    assert str(type_id) == "int[]"


def test_type_id_of_alias():
    MyAlias: TypeAlias = int

    type_id = TypeId.from_type(MyAlias)

    assert type_id == TypeId(type=MyAlias, multi=False) == TypeId(type=int, multi=False)
    assert str(type_id) == "int"


def test_type_id_of_type_of_class():
    class Foo: ...

    type_id = TypeId.from_type(type[Foo])
    assert type_id == TypeId(
        type=type[Foo],
        multi=False,
    )
    assert str(type_id) == "type[Foo]"


def test_type_id_of_callable():
    type_id = TypeId.from_type(Callable[[int], str])
    assert type_id == TypeId(
        type=Callable[[int], str],
        multi=False,
    )
    assert str(type_id) == "Callable[[int], str]"


def test_type_id_of_multi_callable():
    type_id = TypeId.from_type(list[Callable[[int], str]])
    assert type_id == TypeId(
        type=Callable[[int], str],
        multi=True,
    )
    assert str(type_id) == "Callable[[int], str][]"


def test_type_id_of_annotation():
    type_id = TypeId.from_type(Annotated[int, "Activity"])
    assert type_id == TypeId(
        type=Annotated[int, "Activity"],
        multi=False,
    )
    assert str(type_id) == "Annotated[int, Activity]"


def test_type_id_of_complex_annotation():
    type_id = TypeId.from_type(Annotated[Callable[[int], str], "Activity"])
    assert type_id == TypeId(
        type=Annotated[Callable[[int], str], "Activity"],
        multi=False,
    )
    assert str(type_id) == "Annotated[Callable[[int], str], Activity]"


def test_type_id_of_union():
    type_id = TypeId.from_type(int | str)
    assert type_id == TypeId(
        type=int | str,
        multi=False,
    )
    assert str(type_id) == "Union[int, str]"



from abc import ABC, abstractmethod
from typing import ClassVar

from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import PlainTextResponse, Response
from starlette.types import Receive, Scope, Send


class Endpoint(ABC):
    """
    Base class for implementing Starlette endpoints in a way that's compatible with DI, as
    HTTPEndpoint does not allow you to control class initialisation.
    """

    ALLOWED_METHODS: ClassVar[list[str]] = [
        "GET",
        "HEAD",
        "POST",
        "PUT",
        "PATCH",
        "DELETE",
        "OPTIONS",
    ]

    @abstractmethod
    async def exec(self, request: Request) -> Response: ...

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        request = Request(scope, receive=receive)

        if request.method not in self.ALLOWED_METHODS:
            response = await self.method_not_allowed(request)
        else:
            response = await self.exec(request)

        await response(scope, receive, send)

    async def method_not_allowed(self, request: Request) -> Response:
        # If we're running inside a starlette application then raise an
        # exception, so that the configurable exception handler can deal with
        # returning the response. For plain ASGI apps, just return the response.
        headers = {"Allow": ", ".join(self.ALLOWED_METHODS)}
        if "app" in request.scope:
            raise HTTPException(status_code=405, headers=headers)
        return PlainTextResponse("Method Not Allowed", status_code=405, headers=headers)



import time
from typing import Annotated

import pytest
import starlette.testclient
from fastapi import APIRouter, FastAPI
from starlette.websockets import WebSocket

from engin import Engin, Provide, Supply
from engin.extensions.asgi import engin_to_lifespan
from engin.extensions.fastapi import APIRouteDependency, FastAPIEngin, Inject

ROUTER = APIRouter(prefix="")


@ROUTER.get("/")
async def hello_world() -> str:
    return "hello world"


@ROUTER.get("/inject")
async def route_with_dep(some_int: Annotated[int, Inject(int)]) -> int:
    return some_int


@ROUTER.websocket("/websocket")
async def websocket_with_dep(
    websocket: WebSocket, some_int: Annotated[int, Inject(int)]
) -> None:
    await websocket.accept()
    for i in range(5):
        await websocket.send_text(str(i + some_int))
    await websocket.close()


@ROUTER.get("/inject2")
async def route_with_dep_2(
    some_int: Annotated[int, Inject(int)], some_str: Annotated[str, Inject(str)]
) -> int:
    assert some_int == int(some_str)
    return some_int


def app_factory(routers: list[APIRouter]) -> FastAPI:
    app = FastAPI()
    for router in routers:
        app.include_router(router)
    return app


def test_fastapi():
    engin = FastAPIEngin(Provide(app_factory), Supply([ROUTER]))

    with starlette.testclient.TestClient(engin) as client:
        result = client.get("http://127.0.0.1:8000/")

    assert result.status_code == 200
    assert result.json() == "hello world"


def test_inject():
    engin = FastAPIEngin(Provide(app_factory), Supply([ROUTER]), Supply(10))

    with starlette.testclient.TestClient(engin) as client:
        result = client.get("/inject")

    assert result.json() == 10


def test_inject_websocket():
    engin = FastAPIEngin(Provide(app_factory), Supply([ROUTER]), Supply(10))

    with (
        starlette.testclient.TestClient(engin) as client,
        client.websocket_connect("/websocket") as ws,
    ):
        data = ws.receive_text()
        assert data == "10"


async def test_graph():
    engin = FastAPIEngin(Provide(app_factory), Supply([ROUTER]), Supply(10), Supply("a"))

    nodes = engin.graph()

    assert len(nodes) == 8
    assert len([node for node in nodes if isinstance(node.node, APIRouteDependency)]) == 3


async def test_invalid_engin():
    with pytest.raises(LookupError, match="FastAPI"):
        FastAPIEngin()


async def test_engin_to_lifespan():
    engin = Engin(Supply(10))

    app = FastAPI(lifespan=engin_to_lifespan(engin))
    app.include_router(ROUTER)

    with starlette.testclient.TestClient(app) as client:
        result = client.get("http://127.0.0.1:8000/inject")

    assert result.json() == 10


async def test_asgi_request_scope():
    def scoped_factory() -> int:
        return time.time_ns()

    def child_factory(some: int) -> str:
        return str(some)

    app = FastAPIEngin(
        Provide(app_factory),
        Provide(scoped_factory, scope="request"),
        Provide(child_factory),
        Supply([ROUTER]),
    )

    with starlette.testclient.TestClient(app) as client:
        first_call = client.get("http://127.0.0.1:8000/inject").json()
        second_call = client.get("http://127.0.0.1:8000/inject").json()

        # these should differ as the factory is request scoped to each request
        assert first_call != second_call

        client.get("http://127.0.0.1:8000/inject2")



from datetime import UTC, datetime

from engin import Provide, Supply
from engin._graph import DependencyGrapher


def test_dependency_grapher():
    def provide_int() -> int:
        return 3

    def provide_str(num: int) -> str:
        return f"num: {num}"

    def provide_many_str(some: str, dt: datetime) -> list[str]:
        return [some, dt.isoformat()]

    int_provider = Provide(provide_int)
    str_provider = Provide(provide_str)
    many_str_provider = Provide(provide_many_str)
    dt_supplier = Supply(datetime.now(UTC))

    grapher = DependencyGrapher(
        {
            int_provider.return_type_id: int_provider,
            str_provider.return_type_id: str_provider,
            many_str_provider.return_type_id: many_str_provider,
            dt_supplier.return_type_id: dt_supplier,
        }
    )

    nodes = grapher.resolve([many_str_provider])

    assert len(nodes) == 3



import time
from typing import Annotated

import pytest

from engin import Assembler, Entrypoint, Invoke, Provide
from engin.exceptions import NotInScopeError, ProviderError, TypeNotProvidedError
from tests.deps import int_provider, make_many_int, make_many_int_alt, make_str


async def test_assembler():
    assembler = Assembler([int_provider(), Provide(make_str), Provide(make_many_int)])

    def assert_all(some_int: int, some_str: str, many_ints: list[int]):
        assert isinstance(some_str, str)
        assert isinstance(some_int, int)
        assert all(isinstance(x, int) for x in many_ints)

    assembled_dependency = await assembler.assemble(Invoke(assert_all))

    await assembled_dependency()


async def test_assembler_with_multiproviders():
    # catch any non-deterministic ordering bugs
    for _ in range(10):
        assembler = Assembler([Provide(make_many_int), Provide(make_many_int_alt)])

        def assert_all(many_ints: list[int]):
            expected_ints = [*make_many_int(), *make_many_int_alt()]
            assert many_ints == expected_ints

        assembled_dependency = await assembler.assemble(Invoke(assert_all))

        await assembled_dependency()


async def test_assembler_providers_only_called_once():
    _count = 0

    def count() -> int:
        nonlocal _count
        _count += 1
        return _count

    def assert_singleton(some: int) -> None:
        assert some == 1

    assembler = Assembler([Provide(count)])

    assembled_dependency = await assembler.assemble(Invoke(assert_singleton))
    await assembled_dependency()

    assembled_dependency = await assembler.assemble(Invoke(assert_singleton))
    await assembled_dependency()


def test_assembler_with_duplicate_provider_errors():
    with pytest.raises(RuntimeError):
        Assembler([int_provider(), int_provider()])


async def test_assembler_get():
    assembler = Assembler([int_provider(), Provide(make_many_int)])

    assert await assembler.build(int)
    assert await assembler.build(list[int])


async def test_assembler_with_unknown_type_raises_lookup_error():
    assembler = Assembler([])

    with pytest.raises(TypeNotProvidedError):
        await assembler.build(str)

    with pytest.raises(TypeNotProvidedError):
        await assembler.build(list[str])

    with pytest.raises(TypeNotProvidedError):
        await assembler.assemble(Entrypoint(str))


async def test_assembler_with_erroring_provider_raises_provider_error():
    def make_str() -> str:
        raise RuntimeError("foo")

    def make_many_str() -> list[str]:
        raise RuntimeError("foo")

    assembler = Assembler([Provide(make_str), Provide(make_many_str)])

    with pytest.raises(ProviderError):
        await assembler.build(str)

    with pytest.raises(ProviderError):
        await assembler.build(list[str])


async def test_annotations():
    def make_str_1() -> Annotated[str, "1"]:
        return "bar"

    def make_str_2() -> Annotated[str, "2"]:
        return "foo"

    assembler = Assembler([Provide(make_str_1), Provide(make_str_2)])

    with pytest.raises(TypeNotProvidedError):
        await assembler.build(str)

    assert await assembler.build(Annotated[str, "1"]) == "bar"
    assert await assembler.build(Annotated[str, "2"]) == "foo"


async def test_assembler_has():
    def make_str() -> str:
        raise RuntimeError("foo")

    assembler = Assembler([Provide(make_str)])

    assert assembler.has(str)
    assert not assembler.has(int)
    assert not assembler.has(list[str])


async def test_assembler_has_multi():
    def make_str() -> list[str]:
        raise RuntimeError("foo")

    assembler = Assembler([Provide(make_str)])

    assert assembler.has(list[str])
    assert not assembler.has(int)
    assert not assembler.has(str)


async def test_assembler_add():
    assembler = Assembler([])
    assembler.add(int_provider())
    assembler.add(Provide(make_many_int))

    assert assembler.has(int)
    assert assembler.has(list[int])

    # can always add more multiproviders
    assembler.add(Provide(make_many_int))


async def test_assembler_add_overrides():
    def str_provider_a(val: int) -> str:
        return f"a{val}"

    def str_provider_b(val: int) -> str:
        return f"b{val}"

    assembler = Assembler([int_provider(1), Provide(str_provider_a)])

    assert await assembler.build(str) == "a1"

    assembler.add(int_provider(2))
    assembler.add(Provide(str_provider_b))

    assert await assembler.build(str) == "b2"


async def test_assembler_add_clears_caches():
    def make_str(val: int) -> str:
        return str(val)

    assembler = Assembler([int_provider(1), Provide(make_str)])

    assert await assembler.build(int) == 1
    assert await assembler.build(str) == "1"

    assembler.add(int_provider(2))

    assert await assembler.build(int) == 2
    assert await assembler.build(str) == "2"


async def test_assembler_provider_not_in_scope():
    def scoped_provider() -> int:
        return time.time_ns()

    assembler = Assembler([Provide(scoped_provider, scope="foo")])

    with pytest.raises(NotInScopeError):
        await assembler.build(int)


async def test_assembler_provider_scope():
    def scoped_provider() -> int:
        return time.time_ns()

    assembler = Assembler([Provide(scoped_provider, scope="foo")])

    with assembler.scope("foo"):
        await assembler.build(int)

    with pytest.raises(NotInScopeError):
        await assembler.build(int)


async def test_assembler_provider_multi_scope():
    def scoped_provider() -> int:
        return time.time_ns()

    def scoped_provider_2() -> str:
        return "bar"

    assembler = Assembler(
        [Provide(scoped_provider, scope="foo"), Provide(scoped_provider_2, scope="bar")]
    )

    with assembler.scope("foo"):
        await assembler.build(int)
        with assembler.scope("bar"):
            await assembler.build(int)
            await assembler.build(str)
        await assembler.build(int)



from typer.testing import CliRunner

from engin import Engin, Invoke, Provide
from engin._cli._check import cli
from tests.deps import ABlock, make_str

satisfied_engin = Engin(ABlock)


def needs_missing_dependency(missing_type: int) -> None:
    pass


unsatisfied_engin = Engin(
    Provide(make_str),
    Invoke(needs_missing_dependency),
)


def needs_complex_type(custom_type: dict[str, int]) -> None:
    pass


complex_unsatisfied_engin = Engin(
    ABlock,
    Invoke(needs_complex_type),
)

runner = CliRunner()


def test_check_all_dependencies_satisfied():
    result = runner.invoke(
        app=cli,
        args=["tests.cli.test_check:satisfied_engin"],
    )
    assert result.exit_code == 0
    assert "✅ All dependencies are satisfied!" in result.output


def test_check_missing_dependencies():
    result = runner.invoke(
        app=cli,
        args=["tests.cli.test_check:unsatisfied_engin"],
    )
    assert result.exit_code == 1
    assert "❌ Missing providers found:" in result.output
    assert "int" in result.output


def test_check_complex_missing_dependencies():
    result = runner.invoke(
        app=cli,
        args=["tests.cli.test_check:complex_unsatisfied_engin"],
    )
    assert result.exit_code == 1
    assert "❌ Missing providers found:" in result.output
    assert "dict" in result.output


def test_check_invalid_app_path():
    result = runner.invoke(
        app=cli,
        args=["tests.cli.nonexistent:engin"],
    )
    assert result.exit_code == 1
    assert "Unable to find module" in result.output


def test_check_invalid_app_attribute():
    result = runner.invoke(
        app=cli,
        args=["tests.cli.test_check:nonexistent_engin"],
    )
    assert result.exit_code == 1
    assert "has no attribute" in result.output


def test_check_invalid_app_format():
    result = runner.invoke(
        app=cli,
        args=["invalid_format"],
    )
    assert result.exit_code == 1
    assert "Expected an argument of the form 'module:attribute'" in result.output


def test_check_not_engin_instance():
    result = runner.invoke(
        app=cli,
        args=["tests.cli.test_check:runner"],  # CliRunner is not an Engin
    )
    assert result.exit_code == 1
    assert "is not an Engin instance" in result.output



from dataclasses import dataclass

import pytest

from engin._lifecycle import Lifecycle, LifecycleHook


@dataclass
class Tracker:
    state: int = 0

    def start(self) -> None:
        self.state = 1

    def stop(self) -> None:
        self.state = 2


@dataclass
class AsyncTracker:
    state: int = 0

    async def start(self) -> None:
        self.state = 1

    async def stop(self) -> None:
        self.state = 2


@pytest.mark.parametrize("tracker", [(Tracker(), AsyncTracker())])
async def test_lifecycle_hook(tracker):
    tracker = Tracker()

    hook = LifecycleHook(on_start=tracker.start, on_stop=tracker.stop)

    async with hook:
        assert tracker.state == 1

    assert tracker.state == 2


@pytest.mark.parametrize("tracker", [(Tracker(), AsyncTracker())])
async def test_lifecycle_hook_start_only(tracker):
    tracker = Tracker()

    hook = LifecycleHook(on_start=tracker.start)

    async with hook:
        assert tracker.state == 1

    assert tracker.state == 1


@pytest.mark.parametrize("tracker", [(Tracker(), AsyncTracker())])
async def test_lifecycle_hook_stop_only(tracker):
    tracker = Tracker()

    hook = LifecycleHook(on_stop=tracker.stop)

    async with hook:
        assert tracker.state == 0

    assert tracker.state == 2


@pytest.mark.parametrize("tracker", [(Tracker(), AsyncTracker())])
async def test_lifecycle_hook_via_lifecycle(tracker):
    lifecycle = Lifecycle()
    tracker = Tracker()

    lifecycle.hook(on_start=tracker.start, on_stop=tracker.stop)
    cm = lifecycle.list()[0]

    async with cm:
        assert tracker.state == 1

    assert tracker.state == 2


def test_lifecycle_hook_invalid():
    with pytest.raises(ValueError, match="on_start"):
        Lifecycle().hook()



import asyncio
import logging

from httpx import AsyncClient

from engin import Engin, Invoke, Provide

logging.basicConfig(level=logging.DEBUG)


def new_httpx_client() -> AsyncClient:
    return AsyncClient()


async def main(http_client: AsyncClient) -> None:
    res = await http_client.get("https://httpbin.org/get")
    print(res)


engin = Engin(Provide(new_httpx_client), Invoke(main))

if __name__ == "__main__":
    asyncio.run(engin.run())



import tempfile
from pathlib import Path
from unittest.mock import patch

import pytest
from click.exceptions import Exit

from engin import Engin
from engin._cli._common import get_engin_instance

instance_a = Engin()
instance_b = Engin()


def test_get_engin_instance_with_default_from_pyproject():
    pyproject_content = """
[tool.engin]
default-instance = "tests.cli.test_get_engin_instance:instance_a"
"""

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)
        pyproject_path = temp_path / "pyproject.toml"
        pyproject_path.write_text(pyproject_content)

        with patch("pathlib.Path.cwd", return_value=temp_path):
            _, _, found_instance = get_engin_instance()
            assert found_instance == instance_a


def test_get_engin_instance_when_app_overrides_default():
    pyproject_content = """
[tool.engin]
default-instance = "tests.cli.test_check:instance_a"
"""

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)
        pyproject_path = temp_path / "pyproject.toml"
        pyproject_path.write_text(pyproject_content)

        with patch("pathlib.Path.cwd", return_value=temp_path):
            _, _, found_instance = get_engin_instance(
                "tests.cli.test_get_engin_instance:instance_b"
            )
            assert found_instance == instance_b


def test_get_engin_instance_with_no_default_no_app():
    pyproject_content = """
[project]
name = "test"
"""

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)
        pyproject_path = temp_path / "pyproject.toml"
        pyproject_path.write_text(pyproject_content)

        with (
            patch("pathlib.Path.cwd", return_value=temp_path),
            pytest.raises(Exit),
        ):
            get_engin_instance()


def test_get_engin_instance_with_no_default_and_no_pyproject_toml():
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)

        with (
            patch("pathlib.Path.cwd", return_value=temp_path),
            pytest.raises(Exit),
        ):
            get_engin_instance()


def test_check_with_no_default_and_invalid_toml():
    pyproject_content = """
[]]
name -- test
"""

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)
        pyproject_path = temp_path / "pyproject.toml"
        pyproject_path.write_text(pyproject_content)

        with (
            patch("pathlib.Path.cwd", return_value=temp_path),
            pytest.raises(Exit),
        ):
            get_engin_instance()


def test_check_with_no_default_and_invalid_value_type():
    pyproject_content = """
[tool.engin]
default-instance = 3.1
"""

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)
        pyproject_path = temp_path / "pyproject.toml"
        pyproject_path.write_text(pyproject_content)

        with (
            patch("pathlib.Path.cwd", return_value=temp_path),
            pytest.raises(Exit),
        ):
            get_engin_instance()



from engin import Block, provide
from examples.asgi.common.db.adapaters.memory import InMemoryDatabase
from examples.asgi.common.db.ports import DatabaseInterface


class DatabaseBlock(Block):
    @provide
    def database(self) -> DatabaseInterface:
        return InMemoryDatabase()



from typing import Annotated

from fastapi import APIRouter
from pydantic import BaseModel

from engin.extensions.fastapi import Inject
from examples.fastapi.routes.cats.domain import Cat, CatPersonality
from examples.fastapi.routes.cats.ports import CatRepository

router = APIRouter(prefix="/cats")


@router.get("/{cat_id}")
async def get_cat(
    cat_id: int,
    repository: Annotated[CatRepository, Inject(CatRepository)],
) -> Cat:
    return repository.get(cat_id=cat_id)


class CatPostModel(BaseModel):
    name: str
    breed: str
    age: float
    personality: CatPersonality


@router.post("/")
async def post_cat(
    cat: CatPostModel,
    repository: Annotated[CatRepository, Inject(CatRepository)],
) -> int:
    cat_id = repository.next_id()
    cat_domain = Cat(
        id=cat_id,
        name=cat.name,
        personality=cat.personality,
        age=cat.age,
        breed=cat.breed,
    )
    repository.set(cat=cat_domain)
    return cat_id



from unittest.mock import Mock

import pytest

from engin import Block, Engin, invoke, provide
from engin.exceptions import InvalidBlockError


def test_block():
    class MyBlock(Block):
        @provide
        def provide_int(self) -> int:
            return 3

        @invoke
        def invoke_square(self, some: int) -> None: ...

        @provide()
        def provide_str(self) -> str:
            return "3"

        @invoke()
        def invoke_str(self, some: str) -> None: ...

    my_block = MyBlock()

    options = list(my_block._method_options())

    assert len(options) == 4
    assert Engin(my_block)


def test_block_validation_undecorated_method():
    class MyBlock(Block):
        def provide_str(self) -> str:
            return "3"

    with pytest.raises(InvalidBlockError) as exc_info:
        MyBlock.apply(Mock())

    assert "forget" in str(exc_info.value)


def test_block_validation_illegal_decorator():
    class MyBlock(Block):
        def provide_str(self) -> str:
            return "3"

        provide_str._opt = int

    with pytest.raises(InvalidBlockError) as exc_info:
        MyBlock.apply(Mock())

    assert "Invoke" in str(exc_info.value)



import asyncio

import pytest

from engin import Engin, Invoke, Provide
from engin.exceptions import ProviderError


async def test_error_in_provider_when_run():
    async def raise_value_error() -> int:
        raise ValueError("foo")

    async def main(foo: int) -> None:
        return

    engin = Engin(Provide(raise_value_error), Invoke(main))

    with pytest.raises(ProviderError, match="foo"):
        await asyncio.wait_for(engin.run(), timeout=0.5)


async def test_error_in_provider_when_start():
    async def raise_value_error() -> int:
        raise ValueError("foo")

    async def main(foo: int) -> None:
        return

    engin = Engin(Provide(raise_value_error), Invoke(main))

    with pytest.raises(ProviderError, match="foo"):
        await asyncio.wait_for(engin.start(), timeout=0.5)

    # check we can shutdown the app
    await asyncio.wait_for(engin.stop(), timeout=0.5)



from pydantic_settings import BaseSettings
from valkey.asyncio import Valkey

from engin import Block, Lifecycle, provide


class ValkeyConfig(BaseSettings):
    valkey_url: str = "valkey://localhost:6379"


class ValkeyBlock(Block):
    @provide
    def config(self) -> ValkeyConfig:
        return ValkeyConfig()

    @provide
    def client(self, config: ValkeyConfig, lifecycle: Lifecycle) -> Valkey:
        client: Valkey = Valkey.from_url(config.valkey_url)

        # close the client when the app is shutting down
        lifecycle.hook(on_stop=client.aclose)

        return client



import asyncio

import pytest

from engin import Engin, Invoke


async def test_error_in_invocation_when_run():
    async def main() -> None:
        raise ValueError("foo")

    engin = Engin(Invoke(main))

    with pytest.raises(ValueError, match="foo"):
        await asyncio.wait_for(engin.run(), timeout=0.5)


async def test_error_in_invocation_when_start():
    async def main() -> None:
        raise ValueError("foo")

    engin = Engin(Invoke(main))

    with pytest.raises(ValueError, match="foo"):
        await asyncio.wait_for(engin.start(), timeout=0.5)

    # check we can shutdown the app
    await asyncio.wait_for(engin.stop(), timeout=0.5)



from fastapi import APIRouter, FastAPI
from pydantic_settings import BaseSettings

from engin import Block, provide


class AppConfig(BaseSettings):
    debug: bool = False


class AppBlock(Block):
    @provide
    def default_config(self) -> AppConfig:
        return AppConfig()

    @provide
    def app_factory(self, app_config: AppConfig, routers: list[APIRouter]) -> FastAPI:
        app = FastAPI(debug=app_config.debug)

        for router in routers:
            app.include_router(router)

        app.add_api_route(path="/health", endpoint=_health)

        return app


async def _health() -> dict[str, bool]:
    return {"ok": True}



import asyncio
import logging

from engin import Engin, Invoke, Supervisor


async def delayed_error_task():
    raise RuntimeError("Process errored")


def supervise(supervisor: Supervisor) -> None:
    supervisor.supervise(delayed_error_task)


async def test_error_in_supervised_task_handled_when_run(caplog):
    caplog.set_level(logging.INFO)
    engin = Engin(Invoke(supervise))
    await asyncio.wait_for(engin.run(), timeout=0.5)
    assert "Process errored" in caplog.text
    assert engin.is_stopped()


async def test_error_in_supervised_task_handled_when_start(caplog):
    caplog.set_level(logging.INFO)
    engin = Engin(Invoke(supervise))
    await asyncio.wait_for(engin.start(), timeout=0.5)
    await asyncio.sleep(0.1)
    assert "Process errored" in caplog.text
    assert engin.is_stopped()



import asyncio
import signal
import sys

import pytest

from engin import Engin


@pytest.mark.skipif(sys.platform == "win32", reason="`signal.raise_signal` not supported")
async def test_engin_signal_handling_when_run():
    engin = Engin()
    task = asyncio.create_task(engin.run())
    await asyncio.sleep(0.1)
    signal.raise_signal(signal.SIGTERM)
    await asyncio.sleep(0.1)
    assert engin.is_stopped()
    del task


@pytest.mark.skipif(sys.platform == "win32", reason="`signal.raise_signal` not supported")
async def test_engin_signal_handling_when_start():
    engin = Engin()
    await engin.start()
    await asyncio.sleep(0.1)
    signal.raise_signal(signal.SIGTERM)
    await asyncio.sleep(0.1)
    assert engin.is_stopped()



import logging

import uvicorn

from engin import Supply
from engin.extensions.fastapi import FastAPIEngin
from examples.fastapi.app import AppBlock, AppConfig
from examples.fastapi.routes.cats.block import CatBlock

logging.basicConfig(level=logging.DEBUG)

app = FastAPIEngin(AppBlock(), CatBlock(), Supply(AppConfig(debug=True), override=True))


if __name__ == "__main__":
    uvicorn.run(app)



import asyncio
import logging
import random

from valkey.asyncio import Valkey

from engin import Supervisor


class Publisher:
    def __init__(self, valkey: Valkey) -> None:
        self._valkey = valkey

    async def run(self) -> None:
        while True:
            number = random.randint(-100, 100)
            logging.info(f"Publishing: {number}")
            await self._valkey.xadd("numbers", {"number": str(number)})
            await asyncio.sleep(1)


def publisher_factory(valkey: Valkey, supervisor: Supervisor) -> Publisher:
    publisher = Publisher(valkey=valkey)

    # run the publisher as a supervised application task
    supervisor.supervise(publisher.run)

    return publisher



from starlette.routing import Mount, Route

from engin import Block, provide
from examples.asgi.common.db.ports import DatabaseInterface
from examples.asgi.features.cats.api.get import GetCatEndpoint
from examples.asgi.features.cats.api.post import PostCatEndpoint


class CatBlock(Block):
    @provide
    def get_cat_route(self, db: DatabaseInterface) -> GetCatEndpoint:
        return GetCatEndpoint(db=db)

    @provide
    def post_cat_route(self, db: DatabaseInterface) -> PostCatEndpoint:
        return PostCatEndpoint(db=db)

    @provide
    def mount(
        self,
        get_cat_endpoint: GetCatEndpoint,
        post_cat_endpoint: PostCatEndpoint,
    ) -> list[Mount]:
        return [
            Mount(
                "/cats",
                routes=[
                    Route(
                        "/{name:str}",
                        get_cat_endpoint,
                        methods=get_cat_endpoint.ALLOWED_METHODS,
                    ),
                    Route(
                        "/",
                        get_cat_endpoint,
                        methods=get_cat_endpoint.ALLOWED_METHODS,
                    ),
                    Route(
                        "/",
                        post_cat_endpoint,
                        methods=post_cat_endpoint.ALLOWED_METHODS,
                    ),
                ],
            )
        ]



from pydantic_settings import BaseSettings
from starlette.applications import Starlette
from starlette.endpoints import HTTPEndpoint
from starlette.requests import Request
from starlette.responses import JSONResponse, Response
from starlette.routing import Mount, Route

from engin import Block, provide
from engin.extensions.asgi import ASGIType


class AppConfig(BaseSettings):
    debug: bool = False


class HealthCheckEndpoint(HTTPEndpoint):
    async def get(self, _: Request) -> Response:
        return JSONResponse({"ok": True})


class AppBlock(Block):
    @provide
    def app_factory(
        self, routes: list[Route], mounts: list[Mount], app_config: AppConfig
    ) -> ASGIType:
        return Starlette(routes=[*routes, *mounts], debug=app_config.debug)

    @provide
    def default_routes(self) -> list[Route]:
        return [Route("/health", HealthCheckEndpoint)]

    @provide
    def default_config(self) -> AppConfig:
        return AppConfig()



from typing import Any

from examples.asgi.common.db.ports import DatabaseInterface


class InMemoryDatabase(DatabaseInterface):
    def __init__(self) -> None:
        self._data: dict[str, Any] = {}

    def get(self, key: str) -> Any:
        return self._data.get(key)

    def set(self, key: str, value: Any) -> Any:
        self._data[key] = value

    def list(self) -> Any:
        return list(self._data.values())



from typing import Any

import pytest

from engin import Assembler, Provide


class A: ...


class B: ...


class C: ...


class D: ...


class E: ...


class F: ...


class G: ...


def provide_a() -> A:
    return A()


def provide_b(a: A) -> B:
    return B()


def provide_c(a: A, b: B) -> C:
    return C()


def provide_d(a: A, b: B, c: C) -> D:
    return D()


def provide_e(a: A, b: B, c: C, d: D) -> E:
    return E()


def provide_f(a: A, b: B, c: C, d: D, e: E) -> F:
    return F()


def provide_g(a: A, b: B, c: C, d: D, e: E, f: F) -> G:
    return G()


def provide_a_many() -> list[Any]:
    return [A()]


def provide_b_many(a: A) -> list[Any]:
    return [B()]


def provide_c_many(a: A, b: B) -> list[Any]:
    return [C()]


def provide_d_many(a: A, b: B, c: C) -> list[Any]:
    return [D()]


def provide_e_many(a: A, b: B, c: C, d: D) -> list[Any]:
    return [E()]


def provide_f_many(a: A, b: B, c: C, d: D, e: E) -> list[Any]:
    return [F()]


def provide_g_many(a: A, b: B, c: C, d: D, e: E, f: F) -> list[Any]:
    return [G()]


async def main() -> None:
    assembler = Assembler(
        [
            Provide(provide_a),
            Provide(provide_b),
            Provide(provide_c),
            Provide(provide_d),
            Provide(provide_e),
            Provide(provide_f),
            Provide(provide_g),
            Provide(provide_a_many),
            Provide(provide_b_many),
            Provide(provide_c_many),
            Provide(provide_d_many),
            Provide(provide_e_many),
            Provide(provide_f_many),
            Provide(provide_g_many),
        ]
    )
    await assembler.build(G)

    # reset cache
    assembler._assembled_outputs = {}

    await assembler.build(list[Any])


@pytest.mark.benchmark(min_rounds=10000, warmup="on")
def test_assembler_benchmark(aio_benchmark):
    aio_benchmark(main)
